@@ -5,6 +5,7 @@ from itertools import groupby
 from numbers import Rational
 from operator import attrgetter
 from typing import (DefaultDict,
+                    Iterable,
                     List,
                     Optional,
                     Union as Union_)
@@ -24,17 +25,15 @@ from clipping.hints import (Contour,
                             Point,
                             Segment)
 from . import bounding_box
-from .enums import EdgeType
+from .enums import OverlapKind
 from .event import ShapedEvent as Event
 from .events_queue import (BinaryEventsQueue as EventsQueue,
                            BinaryEventsQueueKey as EventsQueueKey)
 from .sweep_line import BinarySweepLine as SweepLine
 from .utils import (all_equal,
-                    contour_to_segments,
-                    sort_pair,
+                    polygon_to_oriented_segments,
                     to_first_boundary_vertex,
                     to_multipolygon_base,
-                    to_multipolygon_contours,
                     to_multipolygon_x_max,
                     to_rational_multipolygon)
 
@@ -70,18 +69,11 @@ class Operation(ABC):
 
     def compute_fields(self, event: Event, below_event: Optional[Event]
                        ) -> None:
-        if below_event is None:
-            event.in_out = False
-            event.other_in_out = True
-        else:
-            if event.from_left is below_event.from_left:
-                event.in_out = not below_event.in_out
-                event.other_in_out = below_event.other_in_out
-            else:
-                event.in_out = not below_event.other_in_out
-                event.other_in_out = (not below_event.in_out
-                                      if below_event.is_vertical
-                                      else below_event.in_out)
+        if below_event is not None:
+            event.other_inside_on_left = (below_event.other_inside_on_left
+                                          if (event.from_left
+                                              is below_event.from_left)
+                                          else below_event.inside_on_left)
             event.below_in_result_event = (below_event.below_in_result_event
                                            if (not self.in_result(below_event)
                                                or below_event.is_vertical)
@@ -113,10 +105,10 @@ class Operation(ABC):
                 end_min, end_max = below_event.complement, event.complement
             if starts_equal:
                 # both line segments are equal or share the left endpoint
-                below_event.edge_type = EdgeType.NON_CONTRIBUTING
-                event.edge_type = (EdgeType.SAME_TRANSITION
-                                   if event.in_out is below_event.in_out
-                                   else EdgeType.DIFFERENT_TRANSITION)
+                below_event.overlap_kind = event.overlap_kind = (
+                    OverlapKind.SAME_ORIENTATION
+                    if event.inside_on_left is below_event.inside_on_left
+                    else OverlapKind.DIFFERENT_ORIENTATION)
                 if not ends_equal:
                     self.divide_segment(end_max.complement, end_min.start)
                 return True
@@ -143,19 +135,20 @@ class Operation(ABC):
         return False
 
     def divide_segment(self, event: Event, point: Point) -> None:
-        left_event = Event(False, point, event.complement, event.from_left)
-        right_event = Event(True, point, event, event.from_left)
+        left_event = Event(False, point, event.complement, event.from_left,
+                           event.inside_on_left)
+        right_event = Event(True, point, event, event.from_left,
+                            event.inside_on_left)
         event.complement.complement, event.complement = left_event, right_event
         self._events_queue.push(left_event)
         self._events_queue.push(right_event)
 
     def fill_queue(self) -> None:
-        for contour in to_multipolygon_contours(self.left):
-            for segment in contour_to_segments(contour):
-                self.register_segment(segment, True)
-        for contour in to_multipolygon_contours(self.right):
-            for segment in contour_to_segments(contour):
-                self.register_segment(segment, False)
+        for polygon in self.left:
+            self.register_segments(polygon_to_oriented_segments(polygon), True)
+        for polygon in self.right:
+            self.register_segments(polygon_to_oriented_segments(polygon),
+                                   False)
 
     @abstractmethod
     def in_result(self, event: Event) -> bool:
@@ -196,13 +189,21 @@ class Operation(ABC):
                 self.compute_fields(below_event, below_below_event)
                 self.compute_fields(event, below_event)
 
-    def register_segment(self, segment: Segment, from_left: bool) -> None:
-        start, end = sort_pair(segment)
-        start_event = Event(False, start, None, from_left)
-        end_event = Event(True, end, start_event, from_left)
-        start_event.complement = end_event
-        self._events_queue.push(start_event)
-        self._events_queue.push(end_event)
+    def register_segments(self,
+                          segments: Iterable[Segment],
+                          from_left: bool) -> None:
+        events_queue = self._events_queue
+        for start, end in segments:
+            inside_on_left = True
+            if start > end:
+                start, end = end, start
+                inside_on_left = False
+            start_event = Event(False, start, None, from_left, inside_on_left)
+            end_event = Event(True, end, start_event, from_left,
+                              inside_on_left)
+            start_event.complement = end_event
+            events_queue.push(start_event)
+            events_queue.push(end_event)
 
     def sweep(self) -> List[Event]:
         self.fill_queue()
@@ -231,10 +232,9 @@ class Difference(Operation):
         return events_to_multipolygon(self.sweep())
 
     def in_result(self, event: Event) -> bool:
-        edge_type = event.edge_type
-        return (edge_type is EdgeType.NORMAL
-                and event.from_left is event.other_in_out
-                or edge_type is EdgeType.DIFFERENT_TRANSITION)
+        return (event.outside
+                if event.from_left
+                else event.inside or event.is_common_polyline_component)
 
     def sweep(self) -> List[Event]:
         self.fill_queue()
@@ -284,12 +284,11 @@ class Intersection(Operation):
         return result
 
     def in_result(self, event: Event) -> bool:
-        edge_type = event.edge_type
-        return (edge_type is EdgeType.NORMAL and not event.other_in_out
-                or edge_type is EdgeType.SAME_TRANSITION)
+        return (event.inside
+                or not event.from_left and event.is_common_region_boundary)
 
 
-class CompleteIntersection(Intersection):
+class CompleteIntersection(Operation):
     __slots__ = ()
 
     def compute(self) -> Mix:
@@ -333,6 +332,24 @@ class CompleteIntersection(Intersection):
         multipolygon = events_to_multipolygon(events)
         return multipoint, multisegment, multipolygon
 
+    def sweep(self) -> List[Event]:
+        self.fill_queue()
+        result = []
+        sweep_line = SweepLine()
+        min_max_x = min(to_multipolygon_x_max(self.left),
+                        to_multipolygon_x_max(self.right))
+        while self._events_queue:
+            event = self._events_queue.pop()
+            start_x, _ = event.start
+            if start_x > min_max_x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
+
+    def in_result(self, event: Event) -> bool:
+        return (event.inside
+                or not event.from_left and event.is_common_region_boundary)
+
 
 class SymmetricDifference(Operation):
     __slots__ = ()
@@ -350,7 +367,7 @@ class SymmetricDifference(Operation):
         return events_to_multipolygon(self.sweep())
 
     def in_result(self, event: Event) -> bool:
-        return event.edge_type is EdgeType.NORMAL
+        return not event.is_overlap
 
 
 class Union(Operation):
@@ -369,9 +386,8 @@ class Union(Operation):
         return events_to_multipolygon(self.sweep())
 
     def in_result(self, event: Event) -> bool:
-        edge_type = event.edge_type
-        return (edge_type is EdgeType.NORMAL and event.other_in_out
-                or edge_type is EdgeType.SAME_TRANSITION)
+        return (event.outside
+                or not event.from_left and event.is_common_region_boundary)
 
 
 def events_to_multipolygon(events: List[Event]) -> Multipolygon:
@@ -421,15 +437,15 @@ def _events_to_contours(events: List[Event],
     depths, parents = defaultdict(int), {}
     processed = [False] * len(events)
     contours = []
+    connectivity = _to_events_connectivity(events)
     for index, event in enumerate(events):
         if processed[index]:
             continue
-
         position = index
         initial = event.start
         contour = [initial]
         steps = [event]
-        while position >= index:
+        while True:
             step = events[position]
             if step.end == initial:
                 break
@@ -438,17 +454,16 @@ def _events_to_contours(events: List[Event],
             position = step.position
             processed[position] = True
             contour.append(events[position].start)
-            position = _to_next_position(position, events, processed, index)
-        position = index if position == -1 else position
+            position = _to_next_position(position, processed, connectivity)
+            if position is None:
+                break
+        position = index if position is None else position
         last_event = events[position]
         processed[position] = processed[last_event.position] = True
-
         _shrink_collinear_vertices(contour)
         if len(contour) < 3:
             continue
-
         contour_id = len(contours)
-
         is_internal = False
         if event.below_in_result_event is not None:
             below_in_result_contour_id = event.below_in_result_event.contour_id
@@ -464,7 +479,6 @@ def _events_to_contours(events: List[Event],
                 depths[contour_id] = depths[below_in_result_contour_id]
                 is_internal = True
         are_internal[contour_id] = is_internal
-
         for step in steps:
             if step.is_right_endpoint:
                 step.complement.result_in_out = True
@@ -474,15 +488,46 @@ def _events_to_contours(events: List[Event],
                 step.contour_id = contour_id
         last_event.complement.result_in_out = True
         last_event.complement.contour_id = contour_id
-
         if depths[contour_id] & 1:
             # holes will be in clockwise order
             contour.reverse()
-
         contours.append(contour)
     return contours
 
 
+def _to_events_connectivity(events: List[Event]) -> List[int]:
+    events_count = len(events)
+    result = [0] * events_count
+    index = 0
+    while index < events_count:
+        current_start = events[index].start
+        right_start_index = index
+        while (index < events_count
+               and events[index].start == current_start
+               and events[index].is_right_endpoint):
+            index += 1
+        right_stop_index = index - 1
+        left_start_index = index
+        while index < events_count and events[index].start == current_start:
+            index += 1
+        left_stop_index = index - 1
+        has_right_events = right_stop_index >= right_start_index
+        has_left_events = left_stop_index >= left_start_index
+        if has_right_events:
+            result[right_start_index:right_stop_index] = range(
+                    right_start_index + 1, right_stop_index + 1)
+            result[right_stop_index] = (left_stop_index
+                                        if has_left_events
+                                        else right_start_index)
+        if has_left_events:
+            result[left_start_index] = (right_start_index
+                                        if has_right_events
+                                        else left_stop_index)
+            result[left_start_index + 1:left_stop_index + 1] = range(
+                    left_start_index, left_stop_index)
+    return result
+
+
 def _shrink_collinear_vertices(contour: Contour) -> None:
     self_intersections, visited = set(), set()
     visit = visited.add
@@ -511,17 +556,12 @@ def _shrink_collinear_vertices(contour: Contour) -> None:
 
 
 def _to_next_position(position: int,
-                      events: List[Event],
                       processed: List[bool],
-                      original_index: int) -> int:
-    result = position + 1
-    point = events[position].start
-    while result < len(events) and events[result].start == point:
-        if not processed[result]:
-            return result
-        else:
-            result += 1
-    result = position - 1
-    while result >= original_index and processed[result]:
-        result -= 1
-    return result
+                      connectivity: List[int]) -> Optional[int]:
+    start_position = position
+    while True:
+        position = connectivity[position]
+        if position == start_position:
+            return None
+        elif not processed[position]:
+            return position
